from copy import deepcopy
from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    TEST_SETTLE_TIMEOUT_MIN,
    ChannelEvent,
    ChannelState,
    MessageTypeId,
)
from raiden_contracts.tests.fixtures.channel import call_settle
from raiden_contracts.tests.utils import (
    EMPTY_ADDITIONAL_HASH,
    EMPTY_BALANCE_HASH,
    EMPTY_SIGNATURE,
    LOCKSROOT_OF_NO_LOCKS,
    UINT256_MAX,
    ChannelValues,
    LockedAmounts,
    call_and_transact,
    fake_bytes,
    get_onchain_settlement_amounts,
    get_settlement_amounts,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.events import check_channel_settled
from raiden_contracts.utils.pending_transfers import (
    get_pending_transfers_tree_with_generated_lists,
)


def test_settle_no_bp_success(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """The simplest settlement without any balance proofs provided"""
    (A, B) = get_accounts(2)
    deposit_A = 10
    deposit_B = 6
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN
    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)

    # Close channel with no balance proof
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    # Do not call updateNonClosingBalanceProof

    # Settlement window must be over before settling the channel
    mine_blocks(web3, settle_timeout + 1)

    # Settling the channel should work with no balance proofs
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier=channel_identifier,
            participant1=A,
            participant1_transferred_amount=0,
            participant1_locked_amount=0,
            participant1_locksroot=LOCKSROOT_OF_NO_LOCKS,
            participant2=B,
            participant2_transferred_amount=0,
            participant2_locked_amount=0,
            participant2_locksroot=LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": A},
    )

    assert custom_token.functions.balanceOf(A).call() == deposit_A
    assert custom_token.functions.balanceOf(B).call() == deposit_B


def test_settle2_no_bp_success(
    web3: Web3,
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """The simplest settlement, tested against the V2 ABI settle"""
    (A, B) = get_accounts(2)
    deposit_A = 10
    deposit_B = 6
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN
    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)

    # Close channel with no balance proof
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    # Do not call updateNonClosingBalanceProof

    # Settlement window must be over before settling the channel
    mine_blocks(web3, settle_timeout + 1)

    # Settling the channel should work with no balance proofs
    call_and_transact(
        token_network.functions.settleChannel2(
            channel_identifier=channel_identifier,
            participant1_settlement=dict(
                participant=A,
                transferred_amount=0,
                locked_amount=0,
                locksroot=LOCKSROOT_OF_NO_LOCKS,
            ),
            participant2_settlement=dict(
                participant=B,
                transferred_amount=0,
                locked_amount=0,
                locksroot=LOCKSROOT_OF_NO_LOCKS,
            ),
        ),
        {"from": A},
    )

    assert custom_token.functions.balanceOf(A).call() == deposit_A
    assert custom_token.functions.balanceOf(B).call() == deposit_B


def test_settle_channel_state(
    web3: Web3,
    get_accounts: Callable,
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    close_and_update_channel: Callable,
    settle_state_tests: Callable,
) -> None:
    """settleChannel() with some balance proofs"""
    (A, B) = get_accounts(2)
    vals_A = ChannelValues(
        deposit=40,
        withdrawn=10,
        transferred=20020,
        locked_amounts=LockedAmounts(claimable_locked=3, unclaimable_locked=4),
    )
    vals_B = ChannelValues(
        deposit=35,
        withdrawn=5,
        transferred=20030,
        locked_amounts=LockedAmounts(claimable_locked=2, unclaimable_locked=3),
    )

    pending_transfers_tree_A = get_pending_transfers_tree_with_generated_lists(
        web3,
        unlockable_amount=vals_A.locked_amounts.claimable_locked,
        expired_amount=vals_A.locked_amounts.unclaimable_locked,
    )
    pending_transfers_tree_B = get_pending_transfers_tree_with_generated_lists(
        web3,
        unlockable_amount=vals_B.locked_amounts.claimable_locked,
        expired_amount=vals_B.locked_amounts.unclaimable_locked,
    )
    vals_A.locksroot = pending_transfers_tree_A.hash_of_packed_transfers
    vals_B.locksroot = pending_transfers_tree_B.hash_of_packed_transfers

    channel_identifier = create_channel_and_deposit(A, B, vals_A.deposit, vals_B.deposit)
    withdraw_channel(channel_identifier, A, vals_A.withdrawn, UINT256_MAX, B)
    withdraw_channel(channel_identifier, B, vals_B.withdrawn, UINT256_MAX, A)
    close_and_update_channel(channel_identifier, A, vals_A, B, vals_B)

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    # Balance & state tests
    settle_state_tests(
        channel_identifier,
        A,
        vals_A,
        B,
        vals_B,
        pre_balance_A,
        pre_balance_B,
        pre_balance_contract,
    )

    # Some manual checks for the final balances, in case the settlement algorithms
    # used in `settle_state_tests` are incorrect

    # FIXME after setTotalWithdraw is implemented again
    post_balance_A = pre_balance_A + 33
    post_balance_B = pre_balance_B + 15
    post_balance_contract = pre_balance_contract - 48

    assert custom_token.functions.balanceOf(A).call() == post_balance_A
    assert custom_token.functions.balanceOf(B).call() == post_balance_B
    assert custom_token.functions.balanceOf(token_network.address).call() == post_balance_contract


def test_settle_single_direct_transfer_for_closing_party(
    web3: Web3,
    get_accounts: Callable,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """Test settle of a channel with one direct transfer to the participant
    that called close.
    """
    (A, B) = get_accounts(2)
    (vals_A, vals_B) = (
        ChannelValues(deposit=1, withdrawn=0, transferred=0),
        ChannelValues(deposit=10, withdrawn=0, transferred=5),
    )
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, vals_A.deposit, B)
    channel_deposit(channel_identifier, B, vals_B.deposit, A)

    balance_proof_B = create_balance_proof(
        channel_identifier,
        B,
        vals_B.transferred,
        vals_B.locked_amounts.locked,
        1,
        LOCKSROOT_OF_NO_LOCKS,
    )
    closing_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), closing_sig_A
        ),
        {"from": A},
    )

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    mine_blocks(web3, settle_timeout + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier=channel_identifier,
            participant1=A,
            participant1_transferred_amount=0,
            participant1_locked_amount=0,
            participant1_locksroot=LOCKSROOT_OF_NO_LOCKS,
            participant2=B,
            participant2_transferred_amount=vals_B.transferred,
            participant2_locked_amount=0,
            participant2_locksroot=LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": A},
    )

    # Calculate how much A and B should receive
    expected_settlement = get_settlement_amounts(vals_A, vals_B)
    # Calculate how much A and B receive according to onchain computation
    onchain_settlement = get_onchain_settlement_amounts(vals_A, vals_B)

    assert expected_settlement.participant1_balance == onchain_settlement.participant1_balance
    assert expected_settlement.participant2_balance == onchain_settlement.participant2_balance
    assert custom_token.functions.balanceOf(A).call() == pre_balance_A + 6
    assert custom_token.functions.balanceOf(B).call() == pre_balance_B + 5
    assert (
        custom_token.functions.balanceOf(token_network.address).call() == pre_balance_contract - 11
    )


def test_settle_single_direct_transfer_for_counterparty(
    web3: Web3,
    get_accounts: Callable,
    custom_token: Contract,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """Test settle of a channel with one direct transfer to the participant
    that did not call close.
    """
    (A, B) = get_accounts(2)
    (vals_A, vals_B) = (
        ChannelValues(deposit=10, withdrawn=0, transferred=5),
        ChannelValues(deposit=1, withdrawn=0, transferred=0),
    )
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, vals_A.deposit, B)
    channel_deposit(channel_identifier, B, vals_B.deposit, A)
    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    balance_proof_A = create_balance_proof(
        channel_identifier,
        A,
        vals_A.transferred,
        vals_A.locked_amounts.locked,
        1,
        LOCKSROOT_OF_NO_LOCKS,
    )

    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )
    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ),
        {"from": B},
    )

    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    mine_blocks(web3, settle_timeout + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier=channel_identifier,
            participant1=B,
            participant1_transferred_amount=0,
            participant1_locked_amount=0,
            participant1_locksroot=LOCKSROOT_OF_NO_LOCKS,
            participant2=A,
            participant2_transferred_amount=vals_A.transferred,
            participant2_locked_amount=0,
            participant2_locksroot=LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": B},
    )

    # Calculate how much A and B should receive
    expected_settlement = get_settlement_amounts(vals_B, vals_A)
    # Calculate how much A and B receive according to onchain computation
    onchain_settlement = get_onchain_settlement_amounts(vals_B, vals_A)

    assert expected_settlement.participant1_balance == onchain_settlement.participant1_balance
    assert expected_settlement.participant2_balance == onchain_settlement.participant2_balance
    assert custom_token.functions.balanceOf(A).call() == pre_balance_A + 5
    assert custom_token.functions.balanceOf(B).call() == pre_balance_B + 6
    assert (
        custom_token.functions.balanceOf(token_network.address).call() == pre_balance_contract - 11
    )


def test_settlement_with_unauthorized_token_transfer(
    web3: Web3,
    get_accounts: Callable,
    custom_token: Contract,
    token_network: Contract,
    assign_tokens: Callable,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    close_and_update_channel: Callable,
) -> None:
    """A participant transfers some tokens to the contract and so loses them"""
    externally_transferred_amount = 5
    (A, B) = get_accounts(2)
    (vals_A, vals_B) = (
        ChannelValues(deposit=35, withdrawn=10, transferred=0),
        ChannelValues(deposit=40, withdrawn=10, transferred=0),
    )
    vals_A.locksroot = fake_bytes(32, "02")
    vals_B.locksroot = fake_bytes(32, "03")

    channel_identifier = create_channel_and_deposit(A, B, vals_A.deposit, vals_B.deposit)

    withdraw_channel(channel_identifier, A, vals_A.withdrawn, UINT256_MAX, B)
    withdraw_channel(channel_identifier, B, vals_B.withdrawn, UINT256_MAX, A)

    close_and_update_channel(channel_identifier, A, vals_A, B, vals_B)

    # Assign additional tokens to A
    assign_tokens(A, externally_transferred_amount)
    assert custom_token.functions.balanceOf(A).call() >= externally_transferred_amount

    # Fetch onchain balances after settlement
    pre_balance_A = custom_token.functions.balanceOf(A).call()
    pre_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # A does a transfer to the token_network without appropriate function call - tokens are lost
    call_and_transact(
        custom_token.functions.transfer(token_network.address, externally_transferred_amount),
        {"from": A},
    )
    assert custom_token.functions.balanceOf(token_network.address).call() == (
        pre_balance_contract + externally_transferred_amount
    )

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)

    # Compute expected settlement amounts
    settlement = get_settlement_amounts(vals_A, vals_B)

    # Channel is settled
    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    # Fetch onchain balances after settlement
    post_balance_A = custom_token.functions.balanceOf(A).call()
    post_balance_B = custom_token.functions.balanceOf(B).call()
    post_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    # A has lost the externally_transferred_amount
    assert (
        pre_balance_A + settlement.participant1_balance - externally_transferred_amount
    ) == post_balance_A

    # B's settlement works correctly
    assert pre_balance_B + settlement.participant2_balance == post_balance_B

    # The externally_transferred_amount stays in the contract
    assert (
        pre_balance_contract
        - settlement.participant1_balance
        - settlement.participant2_balance
        + externally_transferred_amount
    ) == post_balance_contract


def test_settle_wrong_state_fail(
    web3: Web3,
    get_accounts: Callable,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_block: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """settleChannel() fails on OPENED state and on CLOSED state before the settlement block"""
    (A, B) = get_accounts(2)
    vals_A = ChannelValues(deposit=35)
    vals_B = ChannelValues(deposit=40)
    channel_identifier = create_channel_and_deposit(A, B, vals_A.deposit, vals_B.deposit)

    (settle_timeout, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert state == ChannelState.OPENED
    assert settle_timeout == TEST_SETTLE_TIMEOUT_MIN

    with pytest.raises(TransactionFailed, match="TN/settle: channel not closed"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    txn_hash = call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier,
            B,
            A,
            EMPTY_BALANCE_HASH,
            0,
            EMPTY_ADDITIONAL_HASH,
            EMPTY_SIGNATURE,
            closing_sig,
        ),
        {"from": A},
    )

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert state == ChannelState.CLOSED
    assert settle_block_number == TEST_SETTLE_TIMEOUT_MIN + get_block(txn_hash)
    assert web3.eth.block_number < settle_block_number

    with pytest.raises(TransactionFailed, match="TN/settle: settlement timeout"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    assert web3.eth.block_number > settle_block_number

    # Channel is settled
    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)

    (settle_block_number, state) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()
    assert state == ChannelState.REMOVED
    assert settle_block_number == 0


def test_settle_wrong_balance_hash(
    web3: Web3,
    get_accounts: Callable,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    close_and_update_channel: Callable,
    reveal_secrets: Callable,
) -> None:
    """Calling settleChannel() with various wrong arguments and see failures"""
    (A, B) = get_accounts(2)
    vals_A = ChannelValues(
        deposit=35,
        withdrawn=0,
        transferred=5,
        locked_amounts=LockedAmounts(claimable_locked=10, unclaimable_locked=2),
    )
    vals_B = ChannelValues(
        deposit=40,
        withdrawn=0,
        transferred=15,
        locked_amounts=LockedAmounts(claimable_locked=5, unclaimable_locked=4),
    )
    channel_identifier = create_channel_and_deposit(A, B, vals_A.deposit, vals_B.deposit)

    # Mock pending transfers data for A -> B
    pending_transfers_tree_A = get_pending_transfers_tree_with_generated_lists(
        web3,
        unlockable_amount=vals_A.locked_amounts.claimable_locked,
        expired_amount=vals_A.locked_amounts.unclaimable_locked,
    )
    vals_A.locksroot = pending_transfers_tree_A.hash_of_packed_transfers
    # Reveal A's secrets.
    reveal_secrets(A, pending_transfers_tree_A.unlockable)

    # Mock pending transfers data for B -> A
    pending_transfers_tree_B = get_pending_transfers_tree_with_generated_lists(
        web3,
        unlockable_amount=vals_B.locked_amounts.claimable_locked,
        expired_amount=vals_B.locked_amounts.unclaimable_locked,
    )
    vals_B.locksroot = pending_transfers_tree_B.hash_of_packed_transfers
    # Reveal B's secrets
    reveal_secrets(B, pending_transfers_tree_B.unlockable)

    close_and_update_channel(channel_identifier, A, vals_A, B, vals_B)

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)

    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, B, vals_A, A, vals_B)

    vals_A_fail = deepcopy(vals_A)
    vals_A_fail.transferred += 1
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, A, vals_A_fail, B, vals_B)

    vals_A_fail.transferred = 0
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, A, vals_A_fail, B, vals_B)

    vals_A_fail.transferred = UINT256_MAX
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, B, vals_B, A, vals_A_fail)

    vals_A_fail = deepcopy(vals_A)
    vals_A_fail.locked_amounts.claimable_locked += 1
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, A, vals_A_fail, B, vals_B)

    vals_A_fail.locked_amounts.claimable_locked = 0
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, A, vals_A_fail, B, vals_B)

    vals_A_fail.locked_amounts.unclaimable_locked = 0
    vals_A_fail.locked_amounts.claimable_locked = UINT256_MAX
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, B, vals_B, A, vals_A_fail)

    vals_A_fail = deepcopy(vals_A)
    vals_A_fail.locksroot = LOCKSROOT_OF_NO_LOCKS
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, A, vals_A_fail, B, vals_B)

    vals_A_fail.locksroot = fake_bytes(32, "01")
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, A, vals_A_fail, B, vals_B)

    vals_B_fail = deepcopy(vals_B)
    vals_B_fail.transferred += 1
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B_fail)

    vals_B_fail.transferred = 0
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 1"):
        call_settle(token_network, channel_identifier, B, vals_B_fail, A, vals_A)

    vals_B_fail.transferred = UINT256_MAX
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B_fail)

    vals_B_fail = deepcopy(vals_B)
    vals_B_fail.locked_amounts.claimable_locked += 1
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B_fail)

    vals_B_fail.locked_amounts.claimable_locked = 0
    with pytest.raises(AssertionError):
        call_settle(token_network, channel_identifier, B, vals_B_fail, A, vals_A)

    vals_B_fail.locked_amounts.unclaimable_locked = 0
    vals_B_fail.locked_amounts.claimable_locked = UINT256_MAX
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B_fail)

    vals_B_fail = deepcopy(vals_B)
    vals_B_fail.locksroot = LOCKSROOT_OF_NO_LOCKS
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B_fail)

    vals_B_fail.locksroot = fake_bytes(32, "01")
    with pytest.raises(TransactionFailed, match="TN/settle: invalid data for participant 2"):
        call_settle(token_network, channel_identifier, A, vals_A, B, vals_B_fail)

    # Channel is settled
    call_settle(token_network, channel_identifier, A, vals_A, B, vals_B)


def test_settle_channel_event(
    web3: Web3,
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    event_handler: Callable,
) -> None:
    """A successful settleChannel() call causes a SETTLED event"""
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)
    deposit_A = 10
    settle_timeout = TEST_SETTLE_TIMEOUT_MIN

    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)

    balance_proof_A = create_balance_proof(channel_identifier, A, 10, 0, 1, LOCKSROOT_OF_NO_LOCKS)
    balance_proof_B = create_balance_proof(channel_identifier, B, 5, 0, 3, LOCKSROOT_OF_NO_LOCKS)
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        participant=B,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF_UPDATE,
        **balance_proof_A._asdict(),
    )
    close_sig_A = create_balance_proof_countersignature(
        participant=A,
        channel_identifier=channel_identifier,
        msg_type=MessageTypeId.BALANCE_PROOF,
        **balance_proof_B._asdict(),
    )

    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_B._asdict().values(), close_sig_A
        ),
        {"from": A},
    )
    call_and_transact(
        token_network.functions.updateNonClosingBalanceProof(
            channel_identifier,
            A,
            B,
            *balance_proof_A._asdict().values(),
            balance_proof_update_signature_B,
        ),
        {"from": B},
    )

    mine_blocks(web3, settle_timeout + 1)
    txn_hash = call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier=channel_identifier,
            participant1=B,
            participant1_transferred_amount=5,
            participant1_locked_amount=0,
            participant1_locksroot=LOCKSROOT_OF_NO_LOCKS,
            participant2=A,
            participant2_transferred_amount=10,
            participant2_locked_amount=0,
            participant2_locksroot=LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": A},
    )

    ev_handler.add(
        txn_hash,
        ChannelEvent.SETTLED,
        check_channel_settled(
            channel_identifier, B, 5, LOCKSROOT_OF_NO_LOCKS, A, 5, LOCKSROOT_OF_NO_LOCKS
        ),
    )
    ev_handler.check()



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import ValidationError

from raiden_contracts.constants import (
    EMPTY_ADDRESS,
    TEST_SETTLE_TIMEOUT_MIN,
    ChannelEvent,
    ChannelState,
)
from raiden_contracts.tests.utils import (
    EMPTY_ADDITIONAL_HASH,
    EMPTY_BALANCE_HASH,
    EMPTY_SIGNATURE,
    LOCKSROOT_OF_NO_LOCKS,
    UINT256_MAX,
    call_and_transact,
)
from raiden_contracts.tests.utils.blockchain import mine_blocks
from raiden_contracts.utils.events import check_withdraw


def test_withdraw_call(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    """setTotalWithdraw() fails with various wrong arguments"""
    (A, B) = get_accounts(2)
    withdraw_A = 3
    channel_identifier = create_channel_and_deposit(A, B, 10, 1)

    (signature_A_for_A, signature_B_for_A) = create_withdraw_signatures(
        [A, B], channel_identifier, A, withdraw_A, UINT256_MAX
    )

    # Failure with zero (integer) instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=0x0,
            total_withdraw=withdraw_A,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        )

    # Failure with the empty string instead of an address
    with pytest.raises(ValidationError):
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant="",
            total_withdraw=withdraw_A,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        )

    # Failure with a negative number as the total withdrawn amount
    with pytest.raises(ValidationError):
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=A,
            total_withdraw=-1,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        )

    # Failure with an overflown number as the total withdrawn amount
    with pytest.raises(ValidationError):
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=A,
            total_withdraw=UINT256_MAX + 1,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        )

    # Failure with the zero address insted of a participant's address
    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=EMPTY_ADDRESS,
            total_withdraw=withdraw_A,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        ).call({"from": A})

    # Failure with zero as the total withdrawn amount
    with pytest.raises(TransactionFailed, match="TN/withdraw: total withdraw is zero"):
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=A,
            total_withdraw=0,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        ).call({"from": A})

    # Failure with the empty signature instead of A's
    with pytest.raises(TransactionFailed):  # TODO: why doesn't this throw an error message?
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=A,
            total_withdraw=withdraw_A,
            expiration_block=UINT256_MAX,
            participant_signature=EMPTY_SIGNATURE,
            partner_signature=signature_B_for_A,
        ).call({"from": A})

    # Failure with the empty signature instead of B's
    with pytest.raises(TransactionFailed):  # TODO: why doesn't this throw an error message?
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=A,
            total_withdraw=withdraw_A,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=EMPTY_SIGNATURE,
        ).call({"from": A})

    call_and_transact(
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=A,
            total_withdraw=withdraw_A,
            expiration_block=UINT256_MAX,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        ),
        {"from": A},
    )


def test_withdraw_call_near_expiration(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
    web3: Web3,
) -> None:
    """setTotalWithdraw() succeeds when expiration_block is one block in the future"""
    (A, B) = get_accounts(2)
    withdraw_A = 3
    channel_identifier = create_channel_and_deposit(A, B, 10, 1)
    # The block must still be one block in the future when the transaction is
    # processed, so we have to choose an expiration two block in the future
    expiration = web3.eth.block_number + 2

    (signature_A_for_A, signature_B_for_A) = create_withdraw_signatures(
        [A, B], channel_identifier, A, withdraw_A, expiration
    )

    call_and_transact(
        token_network.functions.setTotalWithdraw(
            channel_identifier=channel_identifier,
            participant=A,
            total_withdraw=withdraw_A,
            expiration_block=expiration,
            participant_signature=signature_A_for_A,
            partner_signature=signature_B_for_A,
        ),
        {"from": A},
    )


def test_withdraw_wrong_state(
    web3: Web3,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    withdraw_channel: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """setTotalWithdraw() should fail on a closed or settled channel"""
    (A, B) = get_accounts(2)
    withdraw_A = 1

    assert token_network.functions.getChannelIdentifier(A, B).call() == 0

    channel_identifier = create_channel_and_deposit(A, B, 10, 14, TEST_SETTLE_TIMEOUT_MIN)
    (_, state) = token_network.functions.getChannelInfo(channel_identifier, A, B).call()
    assert state == ChannelState.OPENED

    # Channel is open, withdraw must work
    withdraw_channel(channel_identifier, A, withdraw_A, UINT256_MAX, B)

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier,
            non_closing_participant=B,
            closing_participant=A,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ),
        {"from": A},
    )
    (_, state) = token_network.functions.getChannelInfo(channel_identifier, A, B).call()
    assert state == ChannelState.CLOSED

    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        withdraw_channel(channel_identifier, A, withdraw_A, UINT256_MAX, B)

    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier,
            A,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
            B,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": A},
    )
    (_, state) = token_network.functions.getChannelInfo(channel_identifier, A, B).call()
    assert state == ChannelState.REMOVED

    with pytest.raises(TransactionFailed, match="TN: channel not open"):
        withdraw_channel(channel_identifier, A, withdraw_A, UINT256_MAX, B)


def test_withdraw_bigger(
    create_channel_and_deposit: Callable, get_accounts: Callable, withdraw_channel: Callable
) -> None:
    (A, B) = get_accounts(2)
    deposit_A = 15
    deposit_B = 13

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    with pytest.raises(TransactionFailed, match="TN/withdraw: withdraw > deposit"):
        withdraw_channel(channel_identifier, A, deposit_A + deposit_B + 1, UINT256_MAX, B)
    with pytest.raises(TransactionFailed, match="TN/withdraw: withdraw > deposit"):
        withdraw_channel(channel_identifier, B, deposit_A + deposit_B + 1, UINT256_MAX, A)

    withdraw_channel(channel_identifier, A, 3, UINT256_MAX, B)
    withdraw_channel(channel_identifier, B, 6, UINT256_MAX, A)
    with pytest.raises(TransactionFailed, match="TN/withdraw: withdraw > deposit"):
        withdraw_channel(channel_identifier, A, deposit_A + deposit_B - 5, UINT256_MAX, B)
    with pytest.raises(TransactionFailed, match="TN/withdraw: withdraw > deposit"):
        withdraw_channel(channel_identifier, B, deposit_A + deposit_B - 2, UINT256_MAX, A)

    withdraw_channel(channel_identifier, A, deposit_A + deposit_B - 7, UINT256_MAX, B)


def test_withdraw_wrong_signers(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 15
    deposit_B = 13
    withdraw_A = 5
    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (
        signature_A_for_A,
        signature_B_for_A,
        signature_C_for_A,
    ) = create_withdraw_signatures([A, B, C], channel_identifier, A, withdraw_A, UINT256_MAX)
    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_C_for_A,
            signature_B_for_A,
        ).call({"from": C})
    with pytest.raises(TransactionFailed, match="TN/withdraw: channel id mismatch"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_C_for_A,
        ).call({"from": C})

    call_and_transact(
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_B_for_A,
        ),
        {"from": C},
    )


def test_withdraw_wrong_signature_content(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
    web3: Web3,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 15
    deposit_B = 13
    withdraw_A = 3
    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    channel_identifier_fake = token_network.functions.getChannelIdentifier(A, C).call()

    (signature_A_for_A, signature_B_for_A) = create_withdraw_signatures(
        [A, B], channel_identifier, A, withdraw_A, UINT256_MAX
    )
    (signature_A_for_A_fake1, signature_B_for_A_fake1) = create_withdraw_signatures(
        [A, B], channel_identifier_fake, A, withdraw_A, UINT256_MAX
    )
    (signature_A_for_A_fake2, signature_B_for_A_fake2) = create_withdraw_signatures(
        [A, B], channel_identifier, B, withdraw_A, UINT256_MAX
    )
    (signature_A_for_A_fake3, signature_B_for_A_fake3) = create_withdraw_signatures(
        [A, B], channel_identifier, A, withdraw_A - 1, UINT256_MAX
    )

    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A_fake1,
            signature_B_for_A,
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/withdraw: channel id mismatch"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_B_for_A_fake1,
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A_fake2,
            signature_B_for_A,
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/withdraw: channel id mismatch"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_B_for_A_fake2,
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A_fake3,
            signature_B_for_A,
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/withdraw: channel id mismatch"):
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_B_for_A_fake3,
        ).call({"from": A})
    with pytest.raises(TransactionFailed, match="TN/withdraw: expired"):
        call_and_transact(
            token_network.functions.setTotalWithdraw(
                channel_identifier,
                A,
                withdraw_A,
                0,
                signature_A_for_A,
                signature_B_for_A,
            ),
            {"from": A},
        )
    with pytest.raises(TransactionFailed, match="TN/withdraw: expired"):
        call_and_transact(
            token_network.functions.setTotalWithdraw(
                channel_identifier,
                A,
                withdraw_A,
                web3.eth.block_number,
                signature_A_for_A,
                signature_B_for_A,
            ),
            {"from": A},
        )

    call_and_transact(
        token_network.functions.setTotalWithdraw(
            channel_identifier,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_B_for_A,
        ),
        {"from": A},
    )


def test_withdraw_channel_state(
    get_accounts: Callable,
    token_network: Contract,
    custom_token: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    withdraw_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    withdraw_A = 15
    withdraw_B = 2

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    (
        _,
        withdrawn_amount,
        _,
        _,
        _,
        _,
        _,
    ) = token_network.functions.getChannelParticipantInfo(channel_identifier, A, B).call()
    assert withdrawn_amount == 0

    withdraw_channel(channel_identifier, A, withdraw_A, UINT256_MAX, B, C)

    withdraw_state_tests(
        channel_identifier,
        A,
        deposit_A,
        withdraw_A,
        0,
        balance_A,
        B,
        deposit_B,
        0,
        balance_B,
        balance_contract,
        C,
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    withdraw_channel(channel_identifier, B, withdraw_B, UINT256_MAX, A)

    withdraw_state_tests(
        channel_identifier,
        B,
        deposit_B,
        withdraw_B,
        0,
        balance_B,
        A,
        deposit_A,
        withdraw_A,
        balance_A,
        balance_contract,
    )

    balance_A = custom_token.functions.balanceOf(A).call()
    balance_B = custom_token.functions.balanceOf(B).call()
    balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    withdraw_channel(channel_identifier, B, withdraw_B + 3, UINT256_MAX, A)

    withdraw_state_tests(
        channel_identifier,
        B,
        deposit_B,
        withdraw_B + 3,
        withdraw_B,
        balance_B,
        A,
        deposit_A,
        withdraw_A,
        balance_A,
        balance_contract,
    )


def test_withdraw_replay_reopened_channel(
    web3: Web3,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    get_accounts: Callable,
    create_withdraw_signatures: Callable,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    (A, B) = get_accounts(2)
    deposit_A = 20
    withdraw_A = 5

    channel_identifier1 = create_channel(A, B)[0]
    channel_deposit(channel_identifier1, A, deposit_A, B)
    (signature_A_for_A, signature_B_for_A) = create_withdraw_signatures(
        [A, B], channel_identifier1, A, withdraw_A, UINT256_MAX
    )
    call_and_transact(
        token_network.functions.setTotalWithdraw(
            channel_identifier1,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_B_for_A,
        ),
        {"from": A},
    )

    closing_sig = create_close_signature_for_no_balance_proof(B, channel_identifier1)
    call_and_transact(
        token_network.functions.closeChannel(
            channel_identifier=channel_identifier1,
            non_closing_participant=A,
            closing_participant=B,
            balance_hash=EMPTY_BALANCE_HASH,
            nonce=0,
            additional_hash=EMPTY_ADDITIONAL_HASH,
            non_closing_signature=EMPTY_SIGNATURE,
            closing_signature=closing_sig,
        ),
        {"from": B},
    )
    mine_blocks(web3, TEST_SETTLE_TIMEOUT_MIN + 1)
    call_and_transact(
        token_network.functions.settleChannel(
            channel_identifier1,
            A,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
            B,
            0,
            0,
            LOCKSROOT_OF_NO_LOCKS,
        ),
        {"from": A},
    )

    # Reopen the channel and make sure we cannot use the old withdraw proof
    channel_identifier2 = create_channel(A, B)[0]
    channel_deposit(channel_identifier2, A, deposit_A, B)

    assert channel_identifier1 != channel_identifier2
    with pytest.raises(TransactionFailed, match="TN/withdraw: invalid participant sig"):
        token_network.functions.setTotalWithdraw(
            channel_identifier2,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A,
            signature_B_for_A,
        ).call({"from": A})

    # Signed message with correct channel_identifier must work
    (signature_A_for_A2, signature_B_for_A2) = create_withdraw_signatures(
        [A, B], channel_identifier2, A, withdraw_A, UINT256_MAX
    )
    call_and_transact(
        token_network.functions.setTotalWithdraw(
            channel_identifier2,
            A,
            withdraw_A,
            UINT256_MAX,
            signature_A_for_A2,
            signature_B_for_A2,
        ),
        {"from": A},
    )


def test_withdraw_event(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    withdraw_channel: Callable,
    event_handler: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    ev_handler = event_handler(token_network)

    channel_identifier = create_channel_and_deposit(A, B, 10, 1)

    txn_hash = withdraw_channel(channel_identifier, A, 5, UINT256_MAX, B)
    ev_handler.add(txn_hash, ChannelEvent.WITHDRAW, check_withdraw(channel_identifier, A, 5))

    txn_hash = withdraw_channel(channel_identifier, B, 2, UINT256_MAX, A, C)
    ev_handler.add(txn_hash, ChannelEvent.WITHDRAW, check_withdraw(channel_identifier, B, 2))

    ev_handler.check()



from hashlib import sha256
from typing import Callable

import pytest
from web3.contract import Contract
from web3.exceptions import ValidationError

from raiden_contracts.constants import EVENT_SECRET_REVEALED
from raiden_contracts.tests.utils import call_and_transact
from raiden_contracts.tests.utils.mock import fake_bytes
from raiden_contracts.utils.events import check_secret_revealed, check_secrets_revealed


def test_register_secret_call(secret_registry_contract: Contract) -> None:
    """Test the registrable and not registrable secrets"""
    with pytest.raises(ValidationError):
        secret_registry_contract.functions.registerSecret()
    with pytest.raises(ValidationError):
        secret_registry_contract.functions.registerSecret(3)
    with pytest.raises(ValidationError):
        secret_registry_contract.functions.registerSecret(0)
    with pytest.raises(ValidationError):
        secret_registry_contract.functions.registerSecret("")
    with pytest.raises(ValidationError):
        secret_registry_contract.functions.registerSecret(fake_bytes(33))

    # For interoperability with other SHA256-based hashlocks, even 0x0000..00 needs to be accepted.
    assert secret_registry_contract.functions.registerSecret(fake_bytes(32)).call() is True
    assert secret_registry_contract.functions.registerSecret(fake_bytes(32, "02")).call() is True

    with pytest.raises(ValidationError):
        assert (
            secret_registry_contract.functions.registerSecret(fake_bytes(10, "02")).call() is True
        )


def test_register_secret_return_value(
    secret_registry_contract: Contract, get_accounts: Callable
) -> None:
    """The same secret cannot be registered twice"""
    (A, B) = get_accounts(2)
    secret = b"secretsecretsecretsecretsecretse"

    # We use call here to make sure we would get the correct return value
    # even though this does not change the state
    assert secret_registry_contract.functions.registerSecret(secret).call({"from": A}) is True

    call_and_transact(secret_registry_contract.functions.registerSecret(secret), {"from": A})

    # We use call here to get the return value
    assert secret_registry_contract.functions.registerSecret(secret).call({"from": A}) is False
    assert secret_registry_contract.functions.registerSecret(secret).call({"from": B}) is False


def test_register_secret(
    secret_registry_contract: Contract, get_accounts: Callable, get_block: Callable
) -> None:
    """Register a secret and see it's registered"""
    A = get_accounts(1)[0]
    secret = b"secretsecretsecretsecretsecretse"
    secret2 = b"secretsecretsecretsecretsecretss"
    secrethash = sha256(secret).digest()

    assert secret_registry_contract.functions.getSecretRevealBlockHeight(secrethash).call() == 0

    txn_hash = call_and_transact(
        secret_registry_contract.functions.registerSecret(secret), {"from": A}
    )

    assert secret_registry_contract.functions.getSecretRevealBlockHeight(
        secrethash
    ).call() == get_block(txn_hash)

    # A should be able to register any number of secrets
    call_and_transact(secret_registry_contract.functions.registerSecret(secret2), {"from": A})


def test_register_secret_batch(
    secret_registry_contract: Contract, get_accounts: Callable, get_block: Callable
) -> None:
    """Register four secrets and see them registered"""
    (A,) = get_accounts(1)
    secrets = [fake_bytes(32, fill) for fill in ("02", "03", "04", "05")]
    secret_hashes = [sha256(secret).digest() for secret in secrets]

    for h in secret_hashes:
        assert secret_registry_contract.functions.getSecretRevealBlockHeight(h).call() == 0

    txn_hash = call_and_transact(
        secret_registry_contract.functions.registerSecretBatch(secrets), {"from": A}
    )
    block = get_block(txn_hash)

    for h in secret_hashes:
        assert secret_registry_contract.functions.getSecretRevealBlockHeight(h).call() == block


def test_register_secret_batch_return_value(
    secret_registry_contract: Contract, get_accounts: Callable
) -> None:
    """See registerSecret returns True only when all secrets are registered"""
    (A,) = get_accounts(1)
    secrets = [fake_bytes(32, "02"), fake_bytes(32, "03"), fake_bytes(32)]

    assert secret_registry_contract.functions.registerSecretBatch(secrets).call()

    secrets[2] = fake_bytes(32, "04")
    assert secret_registry_contract.functions.registerSecretBatch(secrets).call() is True

    call_and_transact(secret_registry_contract.functions.registerSecret(secrets[1]), {"from": A})
    assert secret_registry_contract.functions.registerSecretBatch(secrets).call() is False


def test_events(secret_registry_contract: Contract, event_handler: Callable) -> None:
    """A successful registerSecret() call causes an EVENT_SECRET_REVEALED event"""
    secret = b"secretsecretsecretsecretsecretse"
    secrethash = sha256(secret).digest()
    ev_handler = event_handler(secret_registry_contract)

    txn_hash = call_and_transact(secret_registry_contract.functions.registerSecret(secret))

    ev_handler.add(txn_hash, EVENT_SECRET_REVEALED, check_secret_revealed(secrethash, secret))
    ev_handler.check()


def test_register_secret_batch_events(
    secret_registry_contract: Contract, event_handler: Callable
) -> None:
    """A registerSecretBatch() with three secrets causes three EVENT_SECRET_REVEALED events"""
    secrets = [fake_bytes(32, "02"), fake_bytes(32, "03"), fake_bytes(32, "04")]
    secret_hashes = [sha256(secret).digest() for secret in secrets]

    ev_handler = event_handler(secret_registry_contract)

    txn_hash = call_and_transact(secret_registry_contract.functions.registerSecretBatch(secrets))

    ev_handler.add(
        txn_hash,
        EVENT_SECRET_REVEALED,
        check_secrets_revealed(secret_hashes, secrets),
        3,
    )
    ev_handler.check()



import json
from copy import deepcopy
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import IO, Generator, List, Optional
from unittest.mock import MagicMock, patch

import pytest
from click import BadParameter, NoSuchOption
from click.testing import CliRunner
from eth_typing import HexStr
from eth_typing.evm import ChecksumAddress, HexAddress
from eth_utils import ValidationError, to_checksum_address
from pyfakefs.fake_filesystem import FakeFilesystem
from pyfakefs.fake_filesystem_unittest import Patcher
from web3 import Web3
from web3.contract import Contract
from web3.eth import Eth

import raiden_contracts
from raiden_contracts.constants import (
    ALDERAAN_VERSION,
    CONTRACT_MONITORING_SERVICE,
    CONTRACT_ONE_TO_N,
    CONTRACT_SECRET_REGISTRY,
    CONTRACT_SERVICE_REGISTRY,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    CONTRACT_USER_DEPOSIT,
    DEPLOY_SETTLE_TIMEOUT_MAX,
    DEPLOY_SETTLE_TIMEOUT_MIN,
    EMPTY_ADDRESS,
)
from raiden_contracts.contract_manager import DeployedContracts, contracts_precompiled_path
from raiden_contracts.deploy.__main__ import (
    ContractDeployer,
    ContractVerifier,
    contracts_version_with_max_token_networks,
    error_removed_option,
    raiden,
    register,
    services,
    token,
    validate_address,
    verify,
)
from raiden_contracts.deploy.contract_deployer import (
    contracts_version_monitoring_service_takes_token_network_registry,
)
from raiden_contracts.deploy.contract_verifier import (
    _verify_monitoring_service_deployment,
    _verify_user_deposit_deployment,
)
from raiden_contracts.tests.utils import FAKE_ADDRESS, get_random_privkey
from raiden_contracts.tests.utils.constants import (
    DEPLOYER_ADDRESS,
    FAUCET_PRIVATE_KEY,
    SECONDS_PER_DAY,
    SERVICE_DEPOSIT,
    UINT256_MAX,
)
from raiden_contracts.utils.versions import contracts_version_has_initial_service_deposit

GAS_LIMIT = 5860000


@pytest.fixture(scope="session")
def deployer(web3: Web3) -> ContractDeployer:
    return ContractDeployer(
        web3=web3,
        private_key=FAUCET_PRIVATE_KEY,
        gas_limit=GAS_LIMIT,
        gas_price=1,
        wait=10,
        contracts_version=None,
    )


@pytest.fixture(scope="session")
def deployer_0_37_0(web3: Web3) -> ContractDeployer:
    return ContractDeployer(
        web3=web3,
        private_key=FAUCET_PRIVATE_KEY,
        gas_limit=GAS_LIMIT,
        gas_price=1,
        wait=10,
        contracts_version="0.37.0",
    )


@pytest.mark.slow
@pytest.fixture(scope="session")
def deployed_raiden_info(deployer: ContractDeployer) -> DeployedContracts:
    return deployer.deploy_raiden_contracts(
        max_num_of_token_networks=1,
        reuse_secret_registry_from_deploy_file=None,
        settle_timeout_min=DEPLOY_SETTLE_TIMEOUT_MIN,
        settle_timeout_max=DEPLOY_SETTLE_TIMEOUT_MAX,
    )


@pytest.mark.slow
@pytest.fixture(scope="session")
def deployed_raiden_info2(deployer: ContractDeployer) -> DeployedContracts:
    return deployer.deploy_raiden_contracts(
        max_num_of_token_networks=1,
        reuse_secret_registry_from_deploy_file=None,
        settle_timeout_min=DEPLOY_SETTLE_TIMEOUT_MIN,
        settle_timeout_max=DEPLOY_SETTLE_TIMEOUT_MAX,
    )


TOKEN_SUPPLY = 10000000


@pytest.fixture(scope="session")
def token_address(deployer: ContractDeployer) -> HexAddress:
    token_type = "CustomToken"
    deployed_token = deployer.deploy_token_contract(
        token_supply=TOKEN_SUPPLY,
        token_decimals=18,
        token_name="TestToken",
        token_symbol="TTT",
        token_type=token_type,
    )
    return deployed_token[token_type]


DEPOSIT_LIMIT = TOKEN_SUPPLY // 2


@pytest.mark.slow
@pytest.fixture(scope="session")
def deployed_service_info(
    deployer: ContractDeployer,
    token_address: HexAddress,
    token_network_registry_contract: Contract,
) -> DeployedContracts:
    return deployer.deploy_service_contracts(
        token_address=token_address,
        user_deposit_whole_balance_limit=DEPOSIT_LIMIT,
        service_registry_controller=DEPLOYER_ADDRESS,
        initial_service_deposit_price=SERVICE_DEPOSIT // 2,
        service_deposit_bump_numerator=6,
        service_deposit_bump_denominator=5,
        decay_constant=200 * SECONDS_PER_DAY,
        min_price=1000,
        registration_duration=180 * SECONDS_PER_DAY,
        token_network_registry_address=token_network_registry_contract.address,
        reuse_service_registry_from_deploy_file=None,
    )


@pytest.mark.slow
@pytest.fixture(scope="session")
def test_deploy_service_0_37_0(
    deployer_0_37_0: ContractDeployer,
    token_address: HexAddress,
    token_network_registry_contract: Contract,
) -> None:
    with pytest.raises(RuntimeError, match="older service contracts is not supported"):
        deployer_0_37_0.deploy_service_contracts(
            token_address=token_address,
            user_deposit_whole_balance_limit=DEPOSIT_LIMIT,
            service_registry_controller=DEPLOYER_ADDRESS,
            initial_service_deposit_price=SERVICE_DEPOSIT // 2,
            service_deposit_bump_numerator=6,
            service_deposit_bump_denominator=5,
            decay_constant=200 * SECONDS_PER_DAY,
            min_price=1000,
            registration_duration=180 * SECONDS_PER_DAY,
            token_network_registry_address=token_network_registry_contract.address,
            reuse_service_registry_from_deploy_file=None,
        )


@pytest.mark.parametrize(
    "version,expectation",
    [(ALDERAAN_VERSION, True), (None, True)],
)
def test_contracts_version_with_max_token_networks(
    version: Optional[str], expectation: bool
) -> None:
    assert contracts_version_with_max_token_networks(version) == expectation


@pytest.mark.parametrize(
    "version,expectation",
    [(ALDERAAN_VERSION, True), (None, True)],
)
def test_contracts_version_monitoring_service_takes_token_network_registry(
    version: Optional[str], expectation: bool
) -> None:
    assert (
        contracts_version_monitoring_service_takes_token_network_registry(version) == expectation
    )


@pytest.mark.slow
def test_deploy_script_raiden(
    web3: Web3,
    deployer: ContractDeployer,
    deployed_raiden_info: DeployedContracts,
    deployed_raiden_info2: DeployedContracts,
) -> None:
    """Run raiden contracts deployment function and tamper with deployed_contracts_info

    This checks if deploy_raiden_contracts() works correctly in the happy case,
    to make sure no code dependencies have been changed, affecting the deployment script.
    This does not check however that the cli command works correctly.
    This also tampers with deployed_contracts_info to make sure an error is raised in
    verify_deployed_contracts()
    """
    deployed_contracts_info = deployed_raiden_info

    deployer.verify_deployment_data(deployment_data=deployed_contracts_info)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts_version"] = "0.0.0"
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployment_data=deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["chain_id"] = 0
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployment_data=deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_SECRET_REGISTRY]["address"] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_SECRET_REGISTRY]["address"] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY][
        "address"
    ] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_SECRET_REGISTRY]["block_number"] = 0
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_SECRET_REGISTRY]["block_number"] = 0
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY]["block_number"] = 0
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY]["gas_cost"] = 0
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY][
        "address"
    ] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts_version"] = "0.4.0"
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_SECRET_REGISTRY] = deployed_raiden_info2[
        "contracts"
    ][CONTRACT_SECRET_REGISTRY]
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY][
        "constructor_arguments"
    ][0] = DEPLOYER_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY][
        "constructor_arguments"
    ][1] = DEPLOYER_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    deployed_contracts_info_fail = deepcopy(deployed_contracts_info)
    deployed_contracts_info_fail["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY][
        "constructor_arguments"
    ] = []
    with pytest.raises(RuntimeError):
        deployer.verify_deployment_data(deployed_contracts_info_fail)

    # check that it fails if sender has no eth
    deployer = ContractDeployer(
        web3=web3,
        private_key=get_random_privkey(),
        gas_limit=GAS_LIMIT,
        gas_price=1,
        wait=10,
    )
    with pytest.raises(ValidationError):
        deployer.deploy_raiden_contracts(
            1,
            reuse_secret_registry_from_deploy_file=None,
            settle_timeout_min=DEPLOY_SETTLE_TIMEOUT_MIN,
            settle_timeout_max=DEPLOY_SETTLE_TIMEOUT_MAX,
        )


@pytest.mark.slow
def test_deploy_script_raiden_0_37_0(
    deployer_0_37_0: ContractDeployer,
) -> None:
    """Checks that `verify_deployment_data` works fine with older contracts."""
    data = deployer_0_37_0.deploy_raiden_contracts(
        max_num_of_token_networks=1,
        reuse_secret_registry_from_deploy_file=None,
        settle_timeout_min=DEPLOY_SETTLE_TIMEOUT_MIN,
        settle_timeout_max=DEPLOY_SETTLE_TIMEOUT_MAX,
    )

    deployer_0_37_0.verify_deployment_data(deployment_data=data)


def test_deploy_raiden_reuse_secret_registry(
    deployer: ContractDeployer, deployed_raiden_info: DeployedContracts
) -> None:
    """Run deploy_raiden_contracts with a previous SecretRegistry deployment data"""
    with NamedTemporaryFile() as previous_deployment_file:
        previous_deployment_file.write(bytearray(json.dumps(deployed_raiden_info), "ascii"))
        previous_deployment_file.flush()
        new_deployment = deployer.deploy_raiden_contracts(
            1,
            reuse_secret_registry_from_deploy_file=Path(previous_deployment_file.name),
            settle_timeout_min=DEPLOY_SETTLE_TIMEOUT_MIN,
            settle_timeout_max=DEPLOY_SETTLE_TIMEOUT_MAX,
        )
        assert (
            new_deployment["contracts"][CONTRACT_SECRET_REGISTRY]
            == deployed_raiden_info["contracts"][CONTRACT_SECRET_REGISTRY]
        )
        assert (
            new_deployment["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY]
            != deployed_raiden_info["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY]
        )


def test_deploy_services_reuse_service_registry(
    deployer: ContractDeployer,
    deployed_service_info: DeployedContracts,
    token_address: HexAddress,
    token_network_registry_contract: Contract,
) -> None:
    """Run deploy_service_contracts with a previous ServiceRegistry deployment data"""
    with NamedTemporaryFile() as previous_deployment_file:
        previous_deployment_file.write(bytearray(json.dumps(deployed_service_info), "ascii"))
        previous_deployment_file.flush()
        new_deployment = deployer.deploy_service_contracts(
            token_address=token_address,
            user_deposit_whole_balance_limit=DEPOSIT_LIMIT,
            service_registry_controller=DEPLOYER_ADDRESS,
            initial_service_deposit_price=SERVICE_DEPOSIT // 2,
            service_deposit_bump_numerator=6,
            service_deposit_bump_denominator=5,
            decay_constant=200 * SECONDS_PER_DAY,
            min_price=1000,
            registration_duration=180 * SECONDS_PER_DAY,
            token_network_registry_address=token_network_registry_contract.address,
            reuse_service_registry_from_deploy_file=Path(previous_deployment_file.name),
        )
        assert (
            new_deployment["contracts"][CONTRACT_SERVICE_REGISTRY]
            == deployed_service_info["contracts"][CONTRACT_SERVICE_REGISTRY]
        )
        assert (
            new_deployment["contracts"][CONTRACT_ONE_TO_N]
            != deployed_service_info["contracts"][CONTRACT_ONE_TO_N]
        )


def test_deploy_script_token(web3: Web3) -> None:
    """Run test token deployment function used in the deployment script

    This checks if deploy_token_contract() works correctly in the happy case,
    to make sure no code dependencies have been changed, affecting the deployment script.
    This does not check however that the cli command works correctly.
    """
    # normal deployment
    token_type = "CustomToken"
    gas_limit = 5860000
    deployer = ContractDeployer(
        web3=web3,
        private_key=FAUCET_PRIVATE_KEY,
        gas_limit=gas_limit,
        gas_price=1,
        wait=10,
    )

    deployed_token = deployer.deploy_token_contract(
        token_supply=10000000,
        token_decimals=18,
        token_name="TestToken",
        token_symbol="TTT",
        token_type=token_type,
    )

    assert deployed_token[token_type] is not None
    assert isinstance(deployed_token[token_type], str)

    # check that it fails if sender has no eth
    deployer = ContractDeployer(
        web3=web3,
        private_key=get_random_privkey(),
        gas_limit=gas_limit,
        gas_price=1,
        wait=10,
    )
    with pytest.raises(ValidationError):
        deployer.deploy_token_contract(
            token_supply=10000000,
            token_decimals=18,
            token_name="TestToken",
            token_symbol="TTT",
            token_type="CustomToken",
        )


@pytest.mark.slow
def test_deploy_script_register(
    web3: Web3,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    deployed_raiden_info: DeployedContracts,
    token_address: ChecksumAddress,
) -> None:
    """Run token register function used in the deployment script

    This checks if register_token_network() works correctly in the happy case,
    to make sure no code dependencies have been changed, affecting the deployment script.
    This does not check however that the cli command works correctly.
    """
    # normal deployment
    gas_limit = 5860000
    deployer = ContractDeployer(
        web3=web3,
        private_key=FAUCET_PRIVATE_KEY,
        gas_limit=gas_limit,
        gas_price=1,
        wait=10,
    )

    token_registry_abi = deployer.contract_manager.get_contract_abi(
        CONTRACT_TOKEN_NETWORK_REGISTRY
    )
    token_registry_address = deployed_raiden_info["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY][
        "address"
    ]
    with patch(
        "raiden_contracts.deploy.contract_deployer.get_contracts_deployment_info"
    ) as get_deploy_info_mock:
        get_deploy_info_mock.return_value = deployed_raiden_info
        token_network_address = deployer.register_token_network(
            token_registry_abi=token_registry_abi,
            token_registry_address=token_registry_address,
            token_address=token_address,
            channel_participant_deposit_limit=channel_participant_deposit_limit,
            token_network_deposit_limit=token_network_deposit_limit,
        )["token_network_address"]
    assert token_network_address is not None
    assert isinstance(token_network_address, str)


@pytest.mark.slow
def test_deploy_script_service(
    web3: Web3,
    deployed_service_info: DeployedContracts,
    token_address: HexAddress,
    token_network_registry_contract: Contract,
) -> None:
    """Run deploy_service_contracts() used in the deployment script

    This checks if deploy_service_contracts() works correctly in the happy case.
    """
    gas_limit = 5860000
    deployer = ContractDeployer(
        web3=web3,
        private_key=FAUCET_PRIVATE_KEY,
        gas_limit=gas_limit,
        gas_price=1,
        wait=10,
    )

    token_supply = 10000000
    assert isinstance(token_address, str)
    deposit_limit = token_supply // 2

    deployed_service_contracts = deployed_service_info
    deployer.verify_service_contracts_deployment_data(
        token_address=token_address,
        user_deposit_whole_balance_limit=deposit_limit,
        deployed_contracts_info=deployed_service_contracts,
        token_network_registry_address=token_network_registry_contract.address,
    )

    with pytest.raises(RuntimeError):
        assert EMPTY_ADDRESS != token_address
        deployer.verify_service_contracts_deployment_data(
            token_address=EMPTY_ADDRESS,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_service_contracts,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts_version"] = "0.0.0"
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["chain_id"] = deployed_service_contracts["chain_id"] + 1
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_SERVICE_REGISTRY]["constructor_arguments"] = [
        EMPTY_ADDRESS
    ]
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_SERVICE_REGISTRY]["constructor_arguments"][
        0
    ] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    original = deployed_info_fail["contracts"][CONTRACT_USER_DEPOSIT]["constructor_arguments"]
    deployed_info_fail["contracts"][CONTRACT_USER_DEPOSIT]["constructor_arguments"] += original
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_USER_DEPOSIT]["constructor_arguments"][
        0
    ] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_USER_DEPOSIT]["constructor_arguments"][1] = (
        deposit_limit + 1
    )
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    original = deployed_info_fail["contracts"][CONTRACT_MONITORING_SERVICE][
        "constructor_arguments"
    ]
    deployed_info_fail["contracts"][CONTRACT_MONITORING_SERVICE][
        "constructor_arguments"
    ] += original
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_MONITORING_SERVICE]["constructor_arguments"][
        0
    ] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_MONITORING_SERVICE]["constructor_arguments"][
        2
    ] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_MONITORING_SERVICE]["constructor_arguments"][
        3
    ] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_ONE_TO_N]["constructor_arguments"][0] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_ONE_TO_N]["constructor_arguments"][1] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    deployed_info_fail["contracts"][CONTRACT_ONE_TO_N]["constructor_arguments"][2] = EMPTY_ADDRESS
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    deployed_info_fail = deepcopy(deployed_service_contracts)
    original = deployed_info_fail["contracts"][CONTRACT_ONE_TO_N]["constructor_arguments"]
    deployed_info_fail["contracts"][CONTRACT_ONE_TO_N]["constructor_arguments"] += original
    with pytest.raises(RuntimeError):
        deployer.verify_service_contracts_deployment_data(
            token_address=token_address,
            user_deposit_whole_balance_limit=deposit_limit,
            deployed_contracts_info=deployed_info_fail,
            token_network_registry_address=token_network_registry_contract.address,
        )

    def test_missing_deployment(contract_name: str) -> None:
        deployed_info_fail = deepcopy(deployed_service_contracts)
        deployed_info_fail["contracts"][contract_name]["address"] = EMPTY_ADDRESS
        with pytest.raises(RuntimeError):
            deployer.verify_service_contracts_deployment_data(
                token_address=token_address,
                user_deposit_whole_balance_limit=deposit_limit,
                deployed_contracts_info=deployed_info_fail,
                token_network_registry_address=token_network_registry_contract.address,
            )

    for contract_name in [
        CONTRACT_SERVICE_REGISTRY,
        CONTRACT_MONITORING_SERVICE,
        CONTRACT_ONE_TO_N,
        CONTRACT_USER_DEPOSIT,
    ]:
        test_missing_deployment(contract_name)


def test_validate_address_on_none() -> None:
    """validate_address(x, y, None) should return None"""
    mock_command = MagicMock()
    mock_parameter = MagicMock()
    assert validate_address(mock_command, mock_parameter, None) is None


def test_validate_address_empty_string() -> None:
    """validate_address(x, y, '') should return None"""
    assert validate_address(MagicMock(), MagicMock(), "") is None


def test_validate_address_not_an_address() -> None:
    """validate_address(x, y, 'not an address') should raise click.BadParameter"""
    with pytest.raises(BadParameter):
        validate_address(MagicMock(), MagicMock(), "not an address")


def test_validate_address_happy_path() -> None:
    """validate_address(x, y, address) should return the same address checksumed"""
    address = DEPLOYER_ADDRESS
    assert validate_address(MagicMock(), MagicMock(), address) == to_checksum_address(address)


@pytest.fixture
def fs_reload_deployer() -> Generator[FakeFilesystem, None, None]:
    patcher = Patcher(
        modules_to_reload=[
            raiden_contracts.contract_manager,
            raiden_contracts.deploy.__main__,
        ]
    )
    patcher.setUp()
    yield patcher.fs
    patcher.tearDown()


@pytest.mark.slow
def test_store_and_verify_raiden(
    fs_reload_deployer: FakeFilesystem,
    deployed_raiden_info: DeployedContracts,
    deployer: ContractDeployer,
) -> None:
    """Store some raiden contract deployment information and verify them"""
    fs_reload_deployer.add_real_directory(
        contracts_precompiled_path(version=None).parent, read_only=False
    )
    deployed_contracts_info = deployed_raiden_info
    deployer.store_and_verify_deployment_info_raiden(
        deployed_contracts_info=deployed_contracts_info
    )
    deployer.store_and_verify_deployment_info_raiden(
        deployed_contracts_info=deployed_contracts_info
    )


@pytest.mark.slow
def test_store_and_verify_services(
    fs_reload_deployer: FakeFilesystem,
    deployer: ContractDeployer,
    deployed_service_info: DeployedContracts,
    token_address: HexAddress,
    token_network_registry_contract: Contract,
) -> None:
    """Store some service contract deployment information and verify them"""
    fs_reload_deployer.add_real_directory(
        contracts_precompiled_path(version=None).parent, read_only=False
    )
    deployed_contracts_info = deployed_service_info
    deployer.verify_service_contracts_deployment_data(
        token_address=token_address,
        deployed_contracts_info=deployed_contracts_info,
        user_deposit_whole_balance_limit=DEPOSIT_LIMIT,
        token_network_registry_address=token_network_registry_contract.address,
    )
    deployer.store_and_verify_deployment_info_services(
        token_address=token_address,
        deployed_contracts_info=deployed_contracts_info,
        user_deposit_whole_balance_limit=DEPOSIT_LIMIT,
        token_network_registry_address=token_network_registry_contract.address,
    )


@pytest.mark.slow
def test_alderaan_deployer(web3: Web3) -> None:
    """A smoke test for deploying Alderaan version contracts"""
    deployer = ContractDeployer(
        web3=web3,
        private_key=FAUCET_PRIVATE_KEY,
        gas_limit=GAS_LIMIT,
        gas_price=1,
        wait=10,
        contracts_version=ALDERAAN_VERSION,
    )
    deployer.deploy_raiden_contracts(
        max_num_of_token_networks=2,
        reuse_secret_registry_from_deploy_file=None,
        settle_timeout_min=DEPLOY_SETTLE_TIMEOUT_MIN,
        settle_timeout_max=DEPLOY_SETTLE_TIMEOUT_MAX,
    )


def test_error_removed_option_raises() -> None:
    with pytest.raises(NoSuchOption):
        mock = MagicMock()
        error_removed_option("msg")(None, mock, "0xaabbcc")


def test_contracts_version_has_initial_service_deposit() -> None:
    assert contracts_version_has_initial_service_deposit(ALDERAAN_VERSION)
    assert contracts_version_has_initial_service_deposit(None)
    with pytest.raises(ValueError):
        contracts_version_has_initial_service_deposit("not a semver string")


def deploy_token_arguments(privkey: str) -> List[str]:
    return [
        "--rpc-provider",
        "rpc_provider",
        "--private-key",
        privkey,
        "--gas-price",
        "12",
        "--token-supply",
        "20000000",
        "--token-name",
        "ServiceToken",
        "--token-decimals",
        "18",
        "--token-symbol",
        "SVT",
    ]


def test_deploy_token_invalid_privkey() -> None:
    """Call deploy token command with invalid private key"""
    with patch.object(
        ContractDeployer, "deploy_token_contract", spec=ContractDeployer
    ) as mock_deployer:
        runner = CliRunner()
        result = runner.invoke(token, deploy_token_arguments(privkey="wrong_priv_key"))
        assert result.exit_code != 0
        assert type(result.exception) == RuntimeError
        assert result.exception
        assert result.exception.args == ("Could not access the private key.",)
        mock_deployer.assert_not_called()


def test_deploy_token_no_balance(privkey_file: IO) -> None:
    """Call deploy token command with a private key with no balance"""
    with patch.object(
        ContractDeployer, "deploy_token_contract", spec=ContractDeployer
    ) as mock_deployer:
        with patch.object(Eth, "get_balance", return_value=0):
            runner = CliRunner()
            result = runner.invoke(token, deploy_token_arguments(privkey=privkey_file.name))
            assert result.exit_code != 0
            assert type(result.exception) == RuntimeError
            assert result.exception
            assert result.exception.args == ("Account with insufficient funds.",)
            mock_deployer.assert_not_called()


def test_deploy_token_with_balance(privkey_file: IO) -> None:
    """Call deploy token command with a private key with some balance"""
    with patch.object(
        ContractDeployer,
        "deploy_token_contract",
        spec=ContractDeployer,
        return_value={},
    ) as mock_deployer:
        with patch.object(Eth, "get_balance", return_value=100):
            runner = CliRunner()
            result = runner.invoke(token, deploy_token_arguments(privkey=privkey_file.name))
            assert result.exit_code == 0
            mock_deployer.assert_called_once()


def deploy_raiden_arguments(
    privkey: str,
    save_info: Optional[bool],
    contracts_version: Optional[str],
    reuse_secret_registry: bool,
) -> List:
    arguments: List = ["--private-key", privkey, "--rpc-provider", "rpc_provider"]

    if save_info is True:
        arguments.append("--save-info")
    elif save_info is False:
        arguments.append("--no-save-info")

    if contracts_version_with_max_token_networks(contracts_version):
        arguments.extend(["--max-token-networks", 1])

    if contracts_version:
        arguments.extend(["--contracts-version", contracts_version])

    if reuse_secret_registry:
        arguments.extend(["--secret-registry-from-deployment-file", "."])

    return arguments


@patch.object(ContractDeployer, "deploy_raiden_contracts")
@patch.object(ContractVerifier, "store_and_verify_deployment_info_raiden")
@pytest.mark.parametrize("contracts_version", [None, ALDERAAN_VERSION])
@pytest.mark.parametrize("reuse_secret_registry", [False, True])
def test_deploy_raiden(
    mock_deploy: MagicMock,
    mock_verify: MagicMock,
    contracts_version: Optional[str],
    reuse_secret_registry: bool,
    privkey_file: IO,
) -> None:
    """Calling deploy raiden command"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(
            raiden,
            deploy_raiden_arguments(
                privkey=privkey_file.name,
                save_info=None,
                contracts_version=contracts_version,
                reuse_secret_registry=reuse_secret_registry,
            ),
        )
        assert result.exception is None
        assert result.exit_code == 0
        mock_deploy.assert_called_once()
        mock_verify.assert_called_once()


@patch.object(ContractDeployer, "register_token_network")
def test_register_script(
    mock_deploy: MagicMock,
    deployed_raiden_info: DeployedContracts,
    privkey_file: IO,
) -> None:
    """Calling deploy raiden command"""
    with patch(
        "raiden_contracts.deploy.contract_deployer.get_contracts_deployment_info"
    ) as get_deploy_info_mock, patch(
        "raiden_contracts.deploy.__main__._add_token_network_deploy_info"
    ) as add_tn_info:
        get_deploy_info_mock.return_value = deployed_raiden_info
        with patch.object(Eth, "get_balance", return_value=1), patch.object(Eth, "chainId", 61):
            runner = CliRunner()
            result = runner.invoke(
                register,
                [
                    "--rpc-provider",
                    "rpc_provider",
                    "--private-key",
                    privkey_file.name,
                    "--gas-price",
                    "12",
                    "--token-network-registry-address",
                    "0x90a16f6aEA062c429c85dc4124ee4b24A00bCc9a",
                    "--token-address",
                    "0x90a16f6aEA062c429c85dc4124ee4b24A00bCc9a",
                    "--channel-participant-deposit-limit",
                    "100",
                    "--token-network-deposit-limit",
                    "200",
                ],
                catch_exceptions=False,
            )
            assert result.exit_code == 0
            mock_deploy.assert_called_once()
            add_tn_info.assert_called_once()


@patch.object(ContractDeployer, "register_token_network")
def test_register_script_without_token_network(mock_deploy: MagicMock, privkey_file: IO) -> None:
    """Calling deploy raiden command"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(
            register,
            [
                "--rpc-provider",
                "rpv_provider",
                "--private-key",
                privkey_file.name,
                "--gas-price",
                "12",
                "--token-address",
                "0x90a16f6aEA062c429c85dc4124ee4b24A00bCc9a",
                "--channel-participant-deposit-limit",
                "100",
                "--token-network-deposit-limit",
                "200",
            ],
        )
        assert result.exit_code != 0
        assert type(result.exception) == RuntimeError
        assert result.exception
        assert result.exception.args == (
            "No TokenNetworkRegistry was specified. "
            "Add --token-network-registry-address <address>.",
        )
        mock_deploy.assert_not_called()


@patch.object(ContractDeployer, "deploy_raiden_contracts")
@patch.object(ContractDeployer, "verify_deployment_data")
def test_deploy_raiden_save_info_false(
    mock_deploy: MagicMock, mock_verify: MagicMock, privkey_file: IO
) -> None:
    """Calling deploy raiden command with --save_info False"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(
            raiden,
            deploy_raiden_arguments(
                privkey=privkey_file.name,
                save_info=False,
                contracts_version=None,
                reuse_secret_registry=False,
            ),
        )
        assert result.exception is None
        assert result.exit_code == 0
        mock_deploy.assert_called_once()
        mock_verify.assert_called_once()


def deploy_services_arguments(
    privkey: str,
    save_info: Optional[bool],
    service_registry_controller: HexAddress,
    token_network_registry_address: HexAddress,
    contracts_version: Optional[str] = None,
) -> List:
    if save_info is None:
        arguments: List = []
    elif save_info is True:
        arguments = ["--save-info"]
    else:
        arguments = ["--no-save-info"]

    if contracts_version is not None:
        arguments += ["--contracts_version", contracts_version]

    arguments += [
        "--private-key",
        privkey,
        "--rpc-provider",
        "rpc_provider",
        "--user-deposit-whole-limit",
        100,
        "--initial-service-deposit-price",
        SERVICE_DEPOSIT // 2,
        "--service-registry-controller",
        service_registry_controller,
        "--service-deposit-bump-numerator",
        6,
        "--service-deposit-bump-denominator",
        5,
        "--service-deposit-decay-constant",
        200 * SECONDS_PER_DAY,
        "--service-deposit-min-price",
        1000,
        "--service-registration-duration",
        180 * SECONDS_PER_DAY,
        "--token-network-registry-address",
        token_network_registry_address,
    ]
    return arguments


@patch.object(ContractDeployer, "deploy_service_contracts")
@patch.object(ContractVerifier, "store_and_verify_deployment_info_services")
def test_deploy_services(mock_deploy: MagicMock, mock_verify: MagicMock, privkey_file: IO) -> None:
    """Calling deploy raiden command"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(
            services,
            deploy_services_arguments(
                privkey=privkey_file.name,
                save_info=None,
                service_registry_controller=FAKE_ADDRESS,
                token_network_registry_address=FAKE_ADDRESS,
            ),
        )
        assert result.exception is None
        assert result.exit_code == 0
        mock_deploy.assert_called_once()
        mock_verify.assert_called_once()


def test_deploy_old_services(privkey_file: IO) -> None:
    """Calling deploy raiden command"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(
            services,
            deploy_services_arguments(
                privkey=privkey_file.name,
                save_info=None,
                service_registry_controller=FAKE_ADDRESS,
                token_network_registry_address=FAKE_ADDRESS,
                contracts_version="0.21.0",
            ),
        )
        assert result.exception
        assert result.exit_code == 2


@patch.object(ContractDeployer, "deploy_service_contracts")
@patch.object(ContractVerifier, "store_and_verify_deployment_info_services")
def test_deploy_services_with_controller(
    mock_deploy: MagicMock, mock_verify: MagicMock, privkey_file: IO
) -> None:
    """Calling deploy raiden command"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(
            services,
            deploy_services_arguments(
                privkey=privkey_file.name,
                save_info=None,
                service_registry_controller=HexAddress(HexStr("0x" + "1" * 40)),
                token_network_registry_address=FAKE_ADDRESS,
            ),
        )
        assert result.exception is None
        assert result.exit_code == 0
        mock_deploy.assert_called_once()
        mock_verify.assert_called_once()


@patch.object(ContractDeployer, "deploy_service_contracts")
@patch.object(ContractDeployer, "verify_service_contracts_deployment_data")
def test_deploy_services_save_info_false(
    mock_deploy: MagicMock, mock_verify: MagicMock, privkey_file: IO
) -> None:
    """Calling deploy raiden command with --save_info False"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(
            services,
            deploy_services_arguments(
                privkey=privkey_file.name,
                save_info=False,
                service_registry_controller=FAKE_ADDRESS,
                token_network_registry_address=FAKE_ADDRESS,
            ),
        )
        assert result.exception is None
        assert result.exit_code == 0
        mock_deploy.assert_called_once()
        mock_verify.assert_called_once()


@patch.object(ContractVerifier, "verify_deployed_contracts_in_filesystem")
def test_verify_script(mock_verify: MagicMock) -> None:
    """Calling deploy verify command"""
    with patch.object(Eth, "get_balance", return_value=1):
        runner = CliRunner()
        result = runner.invoke(verify, ["--rpc-provider", "rpc-provider"])
        assert result.exception is None
        assert result.exit_code == 0
        mock_verify.assert_called_once()


def test_verify_monitoring_service_deployment_with_wrong_first_constructor_arg(
    token_network_registry_contract: Contract,
) -> None:
    mock_token = MagicMock()
    mock_token.call.return_value = EMPTY_ADDRESS
    mock_monitoring_service = MagicMock()
    mock_monitoring_service.functions.token.return_value = mock_token
    with pytest.raises(RuntimeError):
        _verify_monitoring_service_deployment(
            monitoring_service=mock_monitoring_service,
            constructor_arguments=[FAKE_ADDRESS, 0, 1],
            token_address=EMPTY_ADDRESS,
            service_registry_address=FAKE_ADDRESS,
            user_deposit_address=FAKE_ADDRESS,
            token_network_registry_address=token_network_registry_contract.address,
        )


def test_verify_monitoring_service_deployment_with_wrong_onchain_token_address(
    token_network_registry_contract: Contract,
) -> None:
    mock_token = MagicMock()
    mock_token.call.return_value = EMPTY_ADDRESS
    mock_monitoring_service = MagicMock()
    mock_monitoring_service.functions.token.return_value = mock_token
    with pytest.raises(RuntimeError):
        _verify_monitoring_service_deployment(
            monitoring_service=mock_monitoring_service,
            constructor_arguments=[FAKE_ADDRESS, 0, 1],
            token_address=FAKE_ADDRESS,
            service_registry_address=EMPTY_ADDRESS,
            user_deposit_address=FAKE_ADDRESS,
            token_network_registry_address=token_network_registry_contract.address,
        )


def test_user_deposit_deployment_with_wrong_one_to_n_address() -> None:
    """ContractVerifier.verify_user_deposit_deployment raises on a wrong OneToN address"""
    token_addr = HexAddress(HexStr("0xDa12Dc74D2d0881749CCd9330ac4f0aecda5686a"))
    user_deposit_constructor_arguments = [token_addr, UINT256_MAX]
    wrong_one_to_n_address = FAKE_ADDRESS
    user_deposit_mock = MagicMock()
    mock_token_address = MagicMock()
    mock_token_address.call.return_value = token_addr
    user_deposit_mock.functions.token.return_value = mock_token_address

    with pytest.raises(RuntimeError):
        _verify_user_deposit_deployment(
            user_deposit=user_deposit_mock,
            constructor_arguments=user_deposit_constructor_arguments,
            token_address=token_addr,
            user_deposit_whole_balance_limit=UINT256_MAX,
            one_to_n_address=wrong_one_to_n_address,
            monitoring_service_address=HexAddress(
                HexStr("0xb7765972d78B6C97bB0a5a6b7529DC1fb64aA287")
            ),
        )



from typing import Optional
from unittest.mock import Mock

import pytest
from eth_typing import HexAddress, HexStr
from web3.contract import Contract

from raiden_contracts.constants import (
    ALDERAAN_VERSION,
    CONTRACTS_VERSION,
    EMPTY_ADDRESS,
    DeploymentModule,
)
from raiden_contracts.contract_manager import (
    DeployedContract,
    contracts_data_path,
    contracts_deployed_path,
    get_contracts_deployment_info,
)
from raiden_contracts.deploy.contract_verifier import ContractVerifier
from raiden_contracts.utils.type_aliases import ChainID
from raiden_contracts.utils.versions import contracts_version_provides_services


@pytest.mark.parametrize("version", [None, CONTRACTS_VERSION])
def test_deploy_data_dir_exists(version: Optional[str]) -> None:
    """Make sure directories exist for deployment data"""
    assert contracts_data_path(version).exists(), "deployment data do not exist"
    assert contracts_data_path(version).is_dir()


@pytest.mark.parametrize("version", [None, CONTRACTS_VERSION])
def test_deploy_data_dir_is_not_nested(version: Optional[str]) -> None:
    """Make sure 'data' directories do not contain 'data*' recursively"""
    assert list(contracts_data_path(version).glob("./data*")) == []


@pytest.mark.parametrize("version", [None, CONTRACTS_VERSION])
@pytest.mark.parametrize("chain_id", [3, 4, 5])
@pytest.mark.parametrize("services", [False, True])
def test_deploy_data_file_exists(
    version: Optional[str], chain_id: ChainID, services: bool
) -> None:
    """Make sure files exist for deployment data of each chain_id"""
    assert contracts_deployed_path(chain_id, version, services).exists()


def reasonable_deployment_of_a_contract(deployed: DeployedContract) -> None:
    """Checks an entry under deployment_*.json under a contract name"""
    assert isinstance(deployed["address"], str)
    assert len(deployed["address"]) == 42
    assert isinstance(deployed["transaction_hash"], str)
    assert len(deployed["transaction_hash"]) == 66
    assert isinstance(deployed["block_number"], int)
    assert deployed["block_number"] > 0
    assert isinstance(deployed["gas_cost"], int)
    assert deployed["gas_cost"] > 0
    assert isinstance(deployed["constructor_arguments"], list)


RAIDEN_CONTRACT_NAMES = ("TokenNetworkRegistry", "SecretRegistry")


@pytest.mark.parametrize("version", [None])
@pytest.mark.parametrize("chain_id", [3, 4, 5])
def test_deploy_data_has_fields_raiden(version: Optional[str], chain_id: ChainID) -> None:
    data = get_contracts_deployment_info(chain_id, version, module=DeploymentModule.RAIDEN)
    assert data
    assert data["contracts_version"] == version if version else CONTRACTS_VERSION
    assert data["chain_id"] == chain_id
    contracts = data["contracts"]
    for name in RAIDEN_CONTRACT_NAMES:
        deployed = contracts[name]
        reasonable_deployment_of_a_contract(deployed)


SERVICE_CONTRACT_NAMES = (
    "ServiceRegistry",
    "MonitoringService",
    "OneToN",
    "UserDeposit",
)


@pytest.mark.parametrize("version", [None])
@pytest.mark.parametrize("chain_id", [3, 4, 5])
def test_deploy_data_has_fields_services(version: Optional[str], chain_id: ChainID) -> None:
    data = get_contracts_deployment_info(chain_id, version, module=DeploymentModule.SERVICES)
    assert data
    assert data["contracts_version"] == version if version else CONTRACTS_VERSION
    assert data["chain_id"] == chain_id
    contracts = data["contracts"]
    for name in SERVICE_CONTRACT_NAMES:
        deployed = contracts[name]
        reasonable_deployment_of_a_contract(deployed)


@pytest.mark.parametrize("version", [None])
@pytest.mark.parametrize("chain_id", [3, 4, 5])
def test_deploy_data_all(version: Optional[str], chain_id: ChainID) -> None:
    data_all = get_contracts_deployment_info(chain_id, version, module=DeploymentModule.ALL)
    data_default = get_contracts_deployment_info(chain_id, version)
    assert data_all
    assert data_all == data_default

    for name in RAIDEN_CONTRACT_NAMES + SERVICE_CONTRACT_NAMES:
        deployed = data_all["contracts"][name]
        reasonable_deployment_of_a_contract(deployed)


def test_deploy_data_unknown_module() -> None:
    with pytest.raises(ValueError):
        get_contracts_deployment_info(3, None, module=None)  # type: ignore


def test_deploy_data_not_deployed() -> None:
    assert (
        get_contracts_deployment_info(ChainID(1), "0.8.0", module=DeploymentModule.RAIDEN) is None
    )


@pytest.mark.parametrize("chain_id", [3, 4, 42])
def test_deploy_data_for_redeyes_succeed(chain_id: ChainID) -> None:
    """get_contracts_deployment_info() on RedEyes version should return a non-empty dictionary"""
    assert get_contracts_deployment_info(chain_id, "0.4.0")


@pytest.mark.parametrize("chain_id", [3, 4, 5, 42])
def test_service_deploy_data_for_redeyes_fail(chain_id: ChainID) -> None:
    """get_contracts_deployment_info() on RedEyes version should return a non-empty dictionary"""
    with pytest.raises(ValueError):
        assert get_contracts_deployment_info(chain_id, "0.4.0", DeploymentModule.SERVICES)


@pytest.mark.parametrize("chain_id", [1, 3, 4, 5])
def test_deploy_data_for_alderaan_exist(chain_id: ChainID) -> None:
    """get_contracts_deployment_info() on Alderaan version should have data"""
    assert get_contracts_deployment_info(chain_id, ALDERAAN_VERSION)
    assert get_contracts_deployment_info(chain_id, ALDERAAN_VERSION, DeploymentModule.SERVICES)


def test_version_provides_services() -> None:
    assert contracts_version_provides_services(CONTRACTS_VERSION)
    with pytest.raises(ValueError):
        assert contracts_version_provides_services("not a semver")


def test_verify_nonexistent_deployment(
    user_deposit_whole_balance_limit: int, token_network_registry_address: HexAddress
) -> None:
    """Test verify_deployed_contracts_in_filesystem() with a non-existent deployment data."""
    web3_mock = Mock()
    web3_mock.version.network = 42
    # contracts_version 0.37.0 does not contain a kovan deployment.
    verifier = ContractVerifier(web3=web3_mock, contracts_version=ALDERAAN_VERSION)
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_contracts_in_filesystem()
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_service_contracts_in_filesystem(
            token_address=EMPTY_ADDRESS,
            user_deposit_whole_balance_limit=user_deposit_whole_balance_limit,
            token_network_registry_address=token_network_registry_address,
        )


def test_verify_existent_deployment(token_network_registry_contract: Contract) -> None:
    """Test verify_deployed_contracts_in_filesystem() with an existent deployment data

    but with a fake web3 that returns a wrong block number for deployment.
    """
    web3_mock = Mock()
    web3_mock.version.network = 5
    web3_mock.eth.get_transaction_receipt = lambda _: {"blockNumber": 0}
    verifier = ContractVerifier(web3=web3_mock, contracts_version=ALDERAAN_VERSION)
    # The Mock returns a wrong block number, so the comparison fails.
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_contracts_in_filesystem()
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_service_contracts_in_filesystem(
            token_address=HexAddress(HexStr("0x5Fc523e13fBAc2140F056AD7A96De2cC0C4Cc63A")),
            user_deposit_whole_balance_limit=2 ** 256 - 1,
            token_network_registry_address=token_network_registry_contract.address,
        )


def test_verify_existent_deployment_with_wrong_code(
    token_network_registry_contract: Contract,
) -> None:
    """Test verify_deployed_contracts_in_filesystem() with an existent deployment data

    but with a fake web3 that does not return the correct code.
    """
    web3_mock = Mock()
    web3_mock.version.network = 5
    web3_mock.eth.get_transaction_receipt = lambda _: {
        "blockNumber": 10711807,
        "gasUsed": 555366,
        "contractAddress": "0x8Ff327f7ed03cD6Bd5e611E9e404B47d8c9Db81E",
    }
    verifier = ContractVerifier(web3=web3_mock, contracts_version=ALDERAAN_VERSION)
    # The Mock returns a wrong block number, so the comparison fails.
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_contracts_in_filesystem()
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_service_contracts_in_filesystem(
            token_address=HexAddress(HexStr("0x5Fc523e13fBAc2140F056AD7A96De2cC0C4Cc63A")),
            user_deposit_whole_balance_limit=2 ** 256 - 1,
            token_network_registry_address=token_network_registry_contract.address,
        )



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    EMPTY_ADDRESS,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.tests.utils.constants import DEPLOYER_ADDRESS, NOT_ADDRESS
from raiden_contracts.tests.utils.contracts import call_and_transact


def test_constructor_call(
    web3: Web3,
    get_token_network: Callable,
    custom_token: Contract,
    secret_registry_contract: Contract,
    get_accounts: Callable,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> None:
    """Try to deploy TokenNetwork with various wrong arguments"""

    (A, controller) = get_accounts(2)
    chain_id = web3.eth.chain_id
    settle_min = TEST_SETTLE_TIMEOUT_MIN
    settle_max = TEST_SETTLE_TIMEOUT_MAX

    # failure with no arguments
    with pytest.raises(TypeError):
        get_token_network([])

    # failures with integers instead of a Token address
    with pytest.raises(TypeError):
        get_token_network(
            [
                3,
                secret_registry_contract.address,
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )
    with pytest.raises(TypeError):
        get_token_network(
            [
                0,
                secret_registry_contract.address,
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )

    # failures with non-address strings instead of a Token address
    with pytest.raises(TypeError):
        get_token_network(
            [
                "",
                secret_registry_contract.address,
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )
    with pytest.raises(TypeError):
        get_token_network(
            [
                NOT_ADDRESS,
                secret_registry_contract.address,
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )

    # failures with integers instead of a SecretRegistry address
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                3,
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                0,
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )

    # failures with non-address strings instead of a SecretRegistry address
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                "",
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                NOT_ADDRESS,
                chain_id,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )

    # failures with invalid chain_id
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                secret_registry_contract.address,
                "",
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                secret_registry_contract.address,
                -3,
                settle_min,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )

    # failures with invalid settle_min
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                secret_registry_contract.address,
                chain_id,
                "",
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                secret_registry_contract.address,
                chain_id,
                -3,
                settle_max,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )

    # failures with invalid settle_max
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                secret_registry_contract.address,
                chain_id,
                settle_min,
                "",
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )
    with pytest.raises(TypeError):
        get_token_network(
            [
                custom_token.address,
                secret_registry_contract.address,
                chain_id,
                settle_min,
                -3,
                controller,
                channel_participant_deposit_limit,
                token_network_deposit_limit,
            ]
        )

    # failures with Ethereum addresses that don't contain a Token contract
    with pytest.raises(TransactionFailed, match="TN: invalid token address"):
        get_token_network(
            _token_address=EMPTY_ADDRESS,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )
    with pytest.raises(TransactionFailed, match="TN: invalid token"):
        get_token_network(
            _token_address=A,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )
    with pytest.raises(TransactionFailed):
        get_token_network(
            _token_address=secret_registry_contract.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

    # failures with Ethereum addresses that don't contain the SecretRegistry contract
    with pytest.raises(TransactionFailed, match="TN: invalid SR address"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=EMPTY_ADDRESS,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )
    with pytest.raises(TransactionFailed, match="TN: invalid SR contract"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=A,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

    # failure with chain_id zero
    with pytest.raises(TransactionFailed, match="invalid chain id"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=0,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

    # failure with a timeout min and max swapped
    with pytest.raises(TransactionFailed, match="TN: invalid settle timeouts"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MAX,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MIN,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

    # failure with settle_timeout_min being zero
    with pytest.raises(TransactionFailed, match="TN: invalid settle timeout min"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=0,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MIN,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

    # failure with settle_timeout_max being zero
    with pytest.raises(TransactionFailed, match="TN: invalid settle timeouts"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=0,
            _controller=controller,
            _channel_participant_deposit_limit=channel_participant_deposit_limit,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

    # failure with channel_participant_deposit_limit being zero
    with pytest.raises(TransactionFailed, match="TN: invalid participant limit"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=0,
            _token_network_deposit_limit=token_network_deposit_limit,
        )

    # failure with both limits being zero
    with pytest.raises(TransactionFailed, match="TN: invalid participant limit"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=0,
            _token_network_deposit_limit=0,
        )

    # failure with channel_participant_deposit_limit being bigger than
    # token_network_deposit_limit.
    with pytest.raises(TransactionFailed, match="TN: invalid deposit limits"):
        get_token_network(
            _token_address=custom_token.address,
            _secret_registry=secret_registry_contract.address,
            _chain_id=chain_id,
            _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
            _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
            _controller=controller,
            _channel_participant_deposit_limit=token_network_deposit_limit,
            _token_network_deposit_limit=channel_participant_deposit_limit,
        )

    # see a success to make sure that the above failures are meaningful
    get_token_network(
        _token_address=custom_token.address,
        _secret_registry=secret_registry_contract.address,
        _chain_id=chain_id,
        _settlement_timeout_min=TEST_SETTLE_TIMEOUT_MIN,
        _settlement_timeout_max=TEST_SETTLE_TIMEOUT_MAX,
        _controller=controller,
        _channel_participant_deposit_limit=channel_participant_deposit_limit,
        _token_network_deposit_limit=token_network_deposit_limit,
    )


def test_token_network_variables(token_network: Contract) -> None:
    """Check values of storage variables of the TokenNetwork contract"""
    assert token_network.functions.channel_counter().call() == 0
    assert token_network.functions.signature_prefix().call() == "\x19Ethereum Signed Message:\n"


@pytest.mark.usefixtures("no_token_network")
def test_constructor_not_registered(
    custom_token: Contract,
    secret_registry_contract: Contract,
    token_network_registry_contract: Contract,
    token_network_external: Contract,
) -> None:
    """Check that the TokenNetwork refers to the right Token address and chain_id"""

    token_network = token_network_external
    assert token_network.functions.token().call() == custom_token.address
    assert token_network.functions.secret_registry().call() == secret_registry_contract.address
    assert (
        token_network.functions.chain_id().call()
        == token_network_registry_contract.functions.chain_id().call()
    )

    # The TokenNetworkRegistry doesn't know about the TokenNetwork
    assert (
        token_network_registry_contract.functions.token_to_token_networks(
            custom_token.address
        ).call()
        == EMPTY_ADDRESS
    )


def test_change_owner(
    token_network_external: Contract,
    get_accounts: Callable,
) -> None:
    """Address must be allowed to remove limits after if became owner"""
    token_network = token_network_external
    new_controller = get_accounts(1)[0]

    # Must fail when controller is still DEPLOYER_ADDRESS
    with pytest.raises(TransactionFailed, match="Can only be called by controller"):
        call_and_transact(
            token_network.functions.removeLimits(),
            {"from": new_controller},
        )

    # Must succeed after change of controller
    call_and_transact(
        token_network.functions.changeController(new_controller),
        {"from": DEPLOYER_ADDRESS},
    )
    call_and_transact(
        token_network.functions.removeLimits(),
        {"from": new_controller},
    )



# flake8: noqa

from .base import *
from .channel import *
from .contracts import *
from .deposit_fixtures import *
from .monitoring_service import *
from .one_to_n import *
from .secret_registry import *
from .service_registry_fixtures import *
from .test_contracts import *
from .token import *
from .token_network_fixtures import *
from .token_network_registry import *
from .user_deposit import *



from copy import deepcopy

import pytest

from raiden_contracts.contract_manager import merge_deployment_data


def test_merge_deployment_data_identical() -> None:
    """merge_deployment_data() throws ValueError when identical two dictionaries are given"""
    deployment = {
        "contract_version": "0.12.0",
        "contracts": {"TokenNetworkRegistry": "something"},
    }
    with pytest.raises(ValueError):
        merge_deployment_data(deployment, deployment)  # type: ignore


def test_merge_deployment_data_wrong_chain_id() -> None:
    """merge_deployment_data() throws ValueError on two dictionaries with different chain_id's"""
    deployment1 = {
        "chain_id": 1,
        "contracts_version": "0.12.0",
        "contracts": {"TokenNetworkRegistry": "something"},
    }
    deployment2 = deepcopy(deployment1)
    deployment2["chain_id"] = 2
    deployment2["contracts"] = {"SecretRegistry": "something"}
    with pytest.raises(ValueError):
        merge_deployment_data(deployment1, deployment2)  # type: ignore


def test_merge_deployment_data_wrong_contracts_version() -> None:
    """merge_deployment_data() throws ValueError on dicts with different contracts_version's"""
    deployment1 = {
        "chain_id": 1,
        "contracts_version": "0.12.0",
        "contracts": {"TokenNetworkRegistry": "something"},
    }
    deployment2 = deepcopy(deployment1)
    deployment2["contracts_version"] = "0.13.0"
    deployment2["contracts"] = {"SecretRegistry": "something"}
    with pytest.raises(ValueError):
        merge_deployment_data(deployment1, deployment2)  # type: ignore



from typing import Callable

import pytest
from eth_typing.evm import HexAddress
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_USER_DEPOSIT
from raiden_contracts.tests.fixtures.token import CUSTOM_TOKEN_TOTAL_SUPPLY
from raiden_contracts.tests.utils import call_and_transact


@pytest.fixture(scope="session")
def user_deposit_whole_balance_limit() -> int:
    return CUSTOM_TOKEN_TOTAL_SUPPLY // 100


@pytest.fixture(scope="session")
def uninitialized_user_deposit_contract(
    deploy_tester_contract: Callable,
    custom_token: Contract,
    user_deposit_whole_balance_limit: int,
) -> Contract:
    return deploy_tester_contract(
        CONTRACT_USER_DEPOSIT,
        _token_address=custom_token.address,
        _whole_balance_limit=user_deposit_whole_balance_limit,
    )


@pytest.fixture
def user_deposit_contract(
    uninitialized_user_deposit_contract: Contract,
    monitoring_service_external: Contract,
    one_to_n_contract: Contract,
) -> Contract:
    call_and_transact(
        uninitialized_user_deposit_contract.functions.init(
            monitoring_service_external.address, one_to_n_contract.address
        )
    )
    return uninitialized_user_deposit_contract


@pytest.fixture
def udc_transfer_contract(
    deploy_tester_contract: Callable, uninitialized_user_deposit_contract: Contract
) -> Contract:
    return deploy_tester_contract(
        "UDCTransfer", udc_address=uninitialized_user_deposit_contract.address
    )


@pytest.fixture
def deposit_to_udc(user_deposit_contract: Contract, custom_token: Contract) -> Callable:
    def deposit(receiver: HexAddress, amount: int) -> None:
        """Uses UDC's monotonous deposit amount handling

        If you call it twice, only amount2 - amount1 will be deposited. More
        will be mined and approved to keep the implementation simple, though.
        """
        call_and_transact(custom_token.functions.mint(amount), {"from": receiver})
        call_and_transact(
            custom_token.functions.approve(user_deposit_contract.address, amount),
            {"from": receiver},
        )
        call_and_transact(
            user_deposit_contract.functions.deposit(receiver, amount),
            {"from": receiver},
        )

    return deposit
